# Copyright (C) 2021-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import datetime, timezone
from itertools import chain
import os
from random import randint
import uuid

from django.conf import settings
from django.utils.html import escape

from swh.scheduler.model import LastVisitStatus, ListedOrigin, OriginVisitStats
from swh.web.common.utils import reverse
from swh.web.config import SWH_WEB_SERVER_NAME
from swh.web.misc.coverage import deposited_origins, legacy_origins, listed_origins
from swh.web.tests.django_asserts import assert_contains
from swh.web.tests.utils import check_html_get_response, check_http_get_response


def test_coverage_view_no_metrics(client, swh_scheduler):
    """
    Check coverage view can be rendered when scheduler metrics and deposits
    data are not available.
    """
    url = reverse("swh-coverage")
    check_html_get_response(
        client, url, status_code=200, template_used="misc/coverage.html"
    )


def test_coverage_view_with_metrics(client, swh_scheduler, mocker):
    """
    Generate some sample scheduler metrics and some sample deposits
    that will be consumed by the archive coverage view, then check
    the HTML page gets rendered without errors.
    """

    # mock calls to get nixguix origin counts
    mock_archive = mocker.patch("swh.web.misc.coverage.archive")
    mock_archive.lookup_latest_origin_snapshot.return_value = {"id": "some-snapshot"}
    mock_archive.lookup_snapshot_sizes.return_value = {"release": 30095}

    listers = []
    visit_types = ["git", "hg", "svn", "bzr", "svn"]
    for origins in listed_origins["origins"]:
        # create some instances for each lister
        for instance in range(randint(1, 5)):
            lister = swh_scheduler.get_or_create_lister(
                origins["type"], f"instance-{instance}"
            )
            listers.append(lister)
            # record some sample listed origins
            _origins = []
            origin_visit_stats = []
            for i, visit_type in enumerate(visit_types):
                url = str(uuid.uuid4())
                _origins.append(
                    ListedOrigin(
                        lister_id=lister.id,
                        url=url,
                        visit_type=visit_type,
                        extra_loader_arguments={},
                    )
                )
                # set origin visit stats to some origins
                if i % 2 == 0:
                    now = datetime.now(tz=timezone.utc)
                    origin_visit_stats.append(
                        OriginVisitStats(
                            url=url,
                            visit_type=visit_type,
                            last_successful=now,
                            last_visit=now,
                            last_visit_status=LastVisitStatus.successful,
                            last_snapshot=os.urandom(20),
                        )
                    )
            # send origins data to scheduler
            swh_scheduler.record_listed_origins(_origins)
            swh_scheduler.origin_visit_stats_upsert(origin_visit_stats)

    # compute scheduler metrics
    swh_scheduler.update_metrics()

    # add some sample deposits
    deposits = []
    for origins in deposited_origins["origins"]:
        for _ in range(randint(2, 10)):
            deposits.append(
                {
                    "origin_url": f"https://{origins['search_pattern']}/{uuid.uuid4()}",
                    "status": "done",
                }
            )
    get_deposits_list = mocker.patch("swh.web.misc.coverage.get_deposits_list")
    get_deposits_list.return_value = deposits

    # check view gets rendered without errors
    url = reverse("swh-coverage")
    resp = check_html_get_response(
        client, url, status_code=200, template_used="misc/coverage.html"
    )

    # check logos and origins search links are present in the rendered page
    for origins in chain(
        listed_origins["origins"],
        legacy_origins["origins"],
        deposited_origins["origins"],
    ):
        logo_url = f'{settings.STATIC_URL}img/logos/{origins["type"].lower()}.png'
        assert_contains(resp, f'src="{logo_url}"')

        if "instances" in origins:
            for visit_types_ in origins["instances"].values():
                for data in visit_types_.values():
                    if data["count"]:
                        assert_contains(resp, f'<a href="{escape(data["search_url"])}"')
        else:
            for search_url in origins["search_urls"].values():
                assert_contains(resp, f'<a href="{escape(search_url)}"')

    for visit_type in visit_types:
        assert_contains(resp, f"<td>{visit_type}</td>")

    # check request as in production with cache enabled
    check_http_get_response(
        client, url, status_code=200, server_name=SWH_WEB_SERVER_NAME
    )



# Copyright (C) 2018-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import sentry_sdk

from django.conf import settings
from django.contrib.auth.decorators import user_passes_test
from django.core.paginator import Paginator
from django.http import JsonResponse
from django.shortcuts import render

from swh.web.admin.adminurls import admin_route
from swh.web.auth.utils import ADMIN_LIST_DEPOSIT_PERMISSION
from swh.web.common.utils import (
    get_deposits_list,
    parse_swh_deposit_origin,
    parse_swh_metadata_provenance,
)


def _can_list_deposits(user):
    return user.is_staff or user.has_perm(ADMIN_LIST_DEPOSIT_PERMISSION)


@admin_route(r"deposit/", view_name="admin-deposit")
@user_passes_test(_can_list_deposits, login_url=settings.LOGIN_URL)
def _admin_origin_save(request):
    return render(request, "admin/deposit.html")


@admin_route(r"deposit/list/", view_name="admin-deposit-list")
@user_passes_test(_can_list_deposits, login_url=settings.LOGIN_URL)
def _admin_deposit_list(request):
    table_data = {}
    table_data["draw"] = int(request.GET["draw"])
    try:
        deposits = get_deposits_list(request.GET.get("username"))
        deposits_count = len(deposits)
        search_value = request.GET["search[value]"]
        if search_value:
            deposits = [
                d
                for d in deposits
                if any(
                    search_value.lower() in val
                    for val in [str(v).lower() for v in d.values()]
                )
            ]

        exclude_pattern = request.GET.get("excludePattern")
        if exclude_pattern:
            deposits = [
                d
                for d in deposits
                if all(
                    exclude_pattern.lower() not in val
                    for val in [str(v).lower() for v in d.values()]
                )
            ]

        column_order = request.GET["order[0][column]"]
        field_order = request.GET["columns[%s][name]" % column_order]
        order_dir = request.GET["order[0][dir]"]

        deposits = sorted(deposits, key=lambda d: d[field_order] or "")
        if order_dir == "desc":
            deposits = list(reversed(deposits))

        length = int(request.GET["length"])
        page = int(request.GET["start"]) / length + 1
        paginator = Paginator(deposits, length)
        data = paginator.page(page).object_list
        table_data["recordsTotal"] = deposits_count
        table_data["recordsFiltered"] = len(deposits)
        data_list = []
        for d in data:
            data_dict = {
                "id": d["id"],
                "type": d["type"],
                "external_id": d["external_id"],
                "reception_date": d["reception_date"],
                "status": d["status"],
                "status_detail": d["status_detail"],
                "swhid": d["swhid"],
                "swhid_context": d["swhid_context"],
            }
            provenance = None
            raw_metadata = d["raw_metadata"]
            # Try to determine provenance out of the raw metadata
            if raw_metadata and d["type"] == "meta":  # metadata provenance
                provenance = parse_swh_metadata_provenance(d["raw_metadata"])
            elif raw_metadata and d["type"] == "code":
                provenance = parse_swh_deposit_origin(raw_metadata)

            if not provenance and d["origin_url"]:
                provenance = d["origin_url"]

            # Finally, if still not found, we determine uri using the swhid
            if not provenance and d["swhid_context"]:
                # Trying to compute the origin as we did before in the js
                from swh.model.swhids import QualifiedSWHID

                swhid = QualifiedSWHID.from_string(d["swhid_context"])
                provenance = swhid.origin

            data_dict["uri"] = provenance  # could be None

            # This could be large. As this is not displayed yet, drop it to avoid
            # cluttering the data dict
            data_dict.pop("raw_metadata", None)

            data_list.append(data_dict)

        table_data["data"] = data_list

    except Exception as exc:
        sentry_sdk.capture_exception(exc)
        table_data[
            "error"
        ] = "An error occurred while retrieving the list of deposits !"

    return JsonResponse(table_data)



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json

from django.conf.urls import url
from django.db import IntegrityError
from django.db.models import Q
from django.http.response import (
    HttpResponse,
    HttpResponseBadRequest,
    HttpResponseForbidden,
    HttpResponseNotFound,
)
from rest_framework import serializers
from rest_framework.decorators import api_view
from rest_framework.request import Request
from rest_framework.response import Response

from swh.web.auth.models import UserMailmap, UserMailmapEvent
from swh.web.auth.utils import MAILMAP_PERMISSION


class UserMailmapSerializer(serializers.ModelSerializer):
    class Meta:
        model = UserMailmap
        fields = "__all__"


@api_view(["GET"])
def profile_list_mailmap(request: Request) -> HttpResponse:
    if not request.user.has_perm(MAILMAP_PERMISSION):
        return HttpResponseForbidden()

    mms = UserMailmap.objects.filter(user_id=str(request.user.id),).all()
    return Response(UserMailmapSerializer(mms, many=True).data)


@api_view(["POST"])
def profile_add_mailmap(request: Request) -> HttpResponse:
    if not request.user.has_perm(MAILMAP_PERMISSION):
        return HttpResponseForbidden()

    event = UserMailmapEvent.objects.create(
        user_id=str(request.user.id),
        request_type="add",
        request=json.dumps(request.data),
    )

    from_email = request.data.pop("from_email", None)
    if not from_email:
        return HttpResponseBadRequest("'from_email' must be provided and non-empty.")

    try:
        UserMailmap.objects.create(
            user_id=str(request.user.id), from_email=from_email, **request.data
        )
    except IntegrityError as e:
        if "user_mailmap_from_email_key" in e.args[0]:
            return HttpResponseBadRequest("This 'from_email' already exists.")
        else:
            raise

    event.successful = True
    event.save()

    mm = UserMailmap.objects.get(user_id=str(request.user.id), from_email=from_email)
    return Response(UserMailmapSerializer(mm).data)


@api_view(["POST"])
def profile_update_mailmap(request: Request) -> HttpResponse:
    if not request.user.has_perm(MAILMAP_PERMISSION):
        return HttpResponseForbidden()

    event = UserMailmapEvent.objects.create(
        user_id=str(request.user.id),
        request_type="update",
        request=json.dumps(request.data),
    )

    from_email = request.data.pop("from_email", None)
    if not from_email:
        return HttpResponseBadRequest("'from_email' must be provided and non-empty.")

    user_id = str(request.user.id)

    try:
        to_update = (
            UserMailmap.objects.filter(Q(user_id__isnull=True) | Q(user_id=user_id))
            .filter(from_email=from_email)
            .get()
        )
    except UserMailmap.DoesNotExist:
        return HttpResponseNotFound()

    for attr, value in request.data.items():
        setattr(to_update, attr, value)

    to_update.save()

    event.successful = True
    event.save()

    mm = UserMailmap.objects.get(user_id=user_id, from_email=from_email)
    return Response(UserMailmapSerializer(mm).data)


urlpatterns = [
    url(r"^profile/mailmap/list$", profile_list_mailmap, name="profile-mailmap-list",),
    url(r"^profile/mailmap/add$", profile_add_mailmap, name="profile-mailmap-add",),
    url(
        r"^profile/mailmap/update$",
        profile_update_mailmap,
        name="profile-mailmap-update",
    ),
]



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
from io import StringIO
import json
from typing import Dict

from psycopg2.extras import execute_values
import pytest

from django.core.management import call_command
from django.db import transaction

from swh.model.model import Person
from swh.web.auth.models import UserMailmap, UserMailmapEvent
from swh.web.auth.utils import MAILMAP_PERMISSION
from swh.web.common.utils import reverse
from swh.web.tests.utils import (
    check_api_post_response,
    check_http_get_response,
    create_django_permission,
)


@pytest.fixture
def mailmap_user(regular_user):
    regular_user.user_permissions.add(create_django_permission(MAILMAP_PERMISSION))
    return regular_user


@pytest.mark.django_db(transaction=True)
@pytest.mark.parametrize("view_name", ["profile-mailmap-add", "profile-mailmap-update"])
def test_mailmap_endpoints_anonymous_user(api_client, view_name):
    url = reverse(view_name)
    check_api_post_response(api_client, url, status_code=403)


@pytest.mark.django_db(transaction=True)
def test_mailmap_endpoints_user_with_permission(api_client, mailmap_user):
    api_client.force_login(mailmap_user)

    request_data = {"from_email": "bar@example.org", "display_name": "bar"}

    for view_name in ("profile-mailmap-add", "profile-mailmap-update"):
        url = reverse(view_name)
        check_api_post_response(
            api_client, url, data=request_data, status_code=200,
        )

    # FIXME: use check_api_get_responses; currently this crashes without
    # content_type="application/json"
    resp = check_http_get_response(
        api_client,
        reverse("profile-mailmap-list"),
        status_code=200,
        content_type="application/json",
    ).data
    assert len(resp) == 1
    assert resp[0]["from_email"] == "bar@example.org"
    assert resp[0]["display_name"] == "bar"

    events = UserMailmapEvent.objects.order_by("timestamp").all()
    assert len(events) == 2
    assert events[0].request_type == "add"
    assert json.loads(events[0].request) == request_data
    assert events[1].request_type == "update"
    assert json.loads(events[1].request) == request_data


@pytest.mark.django_db(transaction=True)
def test_mailmap_add_duplicate(api_client, mailmap_user):
    api_client.force_login(mailmap_user)

    check_api_post_response(
        api_client,
        reverse("profile-mailmap-add"),
        data={"from_email": "foo@example.org", "display_name": "bar"},
        status_code=200,
    )
    check_api_post_response(
        api_client,
        reverse("profile-mailmap-add"),
        data={"from_email": "foo@example.org", "display_name": "baz"},
        status_code=400,
    )


@pytest.mark.django_db(transaction=True)
def test_mailmap_add_full(api_client, mailmap_user):
    api_client.force_login(mailmap_user)

    request_data = {
        "from_email": "foo@example.org",
        "from_email_verified": True,
        "from_email_verification_request_date": "2021-02-07T14:04:15Z",
        "display_name": "bar",
        "display_name_activated": True,
        "to_email": "bar@example.org",
        "to_email_verified": True,
        "to_email_verification_request_date": "2021-02-07T15:54:59Z",
    }

    check_api_post_response(
        api_client, reverse("profile-mailmap-add"), data=request_data, status_code=200,
    )

    resp = check_http_get_response(
        api_client,
        reverse("profile-mailmap-list"),
        status_code=200,
        content_type="application/json",
    ).data
    assert len(resp) == 1
    assert resp[0].items() >= request_data.items()

    events = UserMailmapEvent.objects.all()
    assert len(events) == 1
    assert events[0].request_type == "add"
    assert json.loads(events[0].request) == request_data
    assert events[0].successful


@pytest.mark.django_db(transaction=True)
def test_mailmap_endpoints_error_response(api_client, mailmap_user):
    api_client.force_login(mailmap_user)

    url = reverse("profile-mailmap-add")
    resp = check_api_post_response(api_client, url, status_code=400)
    assert b"from_email" in resp.content

    url = reverse("profile-mailmap-update")
    resp = check_api_post_response(api_client, url, status_code=400)
    assert b"from_email" in resp.content

    events = UserMailmapEvent.objects.order_by("timestamp").all()
    assert len(events) == 2

    assert events[0].request_type == "add"
    assert json.loads(events[0].request) == {}
    assert not events[0].successful

    assert events[1].request_type == "update"
    assert json.loads(events[1].request) == {}
    assert not events[1].successful


@pytest.mark.django_db(transaction=True)
def test_mailmap_update(api_client, mailmap_user):
    api_client.force_login(mailmap_user)

    before_add = datetime.datetime.now(tz=datetime.timezone.utc)

    check_api_post_response(
        api_client,
        reverse("profile-mailmap-add"),
        data={"from_email": "orig1@example.org", "display_name": "Display Name 1"},
        status_code=200,
    )
    check_api_post_response(
        api_client,
        reverse("profile-mailmap-add"),
        data={"from_email": "orig2@example.org", "display_name": "Display Name 2"},
        status_code=200,
    )
    after_add = datetime.datetime.now(tz=datetime.timezone.utc)

    mailmaps = list(UserMailmap.objects.order_by("from_email").all())
    assert len(mailmaps) == 2, mailmaps

    assert mailmaps[0].from_email == "orig1@example.org", mailmaps
    assert mailmaps[0].display_name == "Display Name 1", mailmaps
    assert before_add <= mailmaps[0].last_update_date <= after_add

    assert mailmaps[1].from_email == "orig2@example.org", mailmaps
    assert mailmaps[1].display_name == "Display Name 2", mailmaps
    assert before_add <= mailmaps[0].last_update_date <= after_add

    before_update = datetime.datetime.now(tz=datetime.timezone.utc)

    check_api_post_response(
        api_client,
        reverse("profile-mailmap-update"),
        data={"from_email": "orig1@example.org", "display_name": "Display Name 1b"},
        status_code=200,
    )

    after_update = datetime.datetime.now(tz=datetime.timezone.utc)

    mailmaps = list(UserMailmap.objects.order_by("from_email").all())
    assert len(mailmaps) == 2, mailmaps

    assert mailmaps[0].from_email == "orig1@example.org", mailmaps
    assert mailmaps[0].display_name == "Display Name 1b", mailmaps
    assert before_update <= mailmaps[0].last_update_date <= after_update

    assert mailmaps[1].from_email == "orig2@example.org", mailmaps
    assert mailmaps[1].display_name == "Display Name 2", mailmaps
    assert before_add <= mailmaps[1].last_update_date <= after_add

    events = UserMailmapEvent.objects.order_by("timestamp").all()
    assert len(events) == 3
    assert events[0].request_type == "add"
    assert events[1].request_type == "add"
    assert events[2].request_type == "update"


def populate_mailmap():
    for (verified, activated) in (
        (False, False),
        (False, True),
        (True, False),
        (True, True),
    ):
        verified_str = "V" if verified else ""
        activated_str = "A" if activated else ""
        UserMailmap.objects.create(
            from_email=f"from_email{verified_str}{activated_str}@example.com",
            display_name=f"Display Name {verified_str} {activated_str}".strip(),
            from_email_verified=verified,
            display_name_activated=activated,
        )


def call_sync_mailmaps(*args) -> str:
    out = StringIO()
    err = StringIO()

    call_command("sync_mailmaps", *args, stdout=out, stderr=err)

    out.seek(0)
    err.seek(0)
    assert err.read() == ""
    return out.read()


MAILMAP_KNOWN_FULLNAMES = (
    "Original Name <from_email@example.com>",
    "Original Name V <from_emailV@example.com>",
    "Original Name A <from_emailA@example.com>",
    "Original Name V A <from_emailVA@example.com>",
    "Original Name V A 2 <from_emailVA@example.com>",
    "Original Name V A 3 <from_emailVA@example.com>",
)


MAILMAP_KNOWN_PEOPLE = tuple(
    Person.from_fullname(f.encode()) for f in MAILMAP_KNOWN_FULLNAMES
)


def init_stub_storage_db(postgresql):
    cur = postgresql.cursor()
    cur.execute(
        """
        CREATE TABLE person (
          fullname bytea PRIMARY KEY,
          name bytea,
          email bytea,
          displayname bytea
        )
        """
    )
    execute_values(
        cur,
        "INSERT INTO person (fullname, name, email) VALUES %s",
        (p.to_dict() for p in MAILMAP_KNOWN_PEOPLE),
        template="(%(fullname)s, %(name)s, %(email)s)",
    )
    cur.execute("CREATE INDEX ON person (email)")
    postgresql.commit()
    cur.close()

    return postgresql.dsn


def get_displaynames(postgresql) -> Dict[str, str]:
    with postgresql.cursor() as cur:
        cur.execute(
            "SELECT fullname, displayname FROM person WHERE displayname IS NOT NULL"
        )

        return {bytes(f).decode("utf-8"): bytes(d).decode("utf-8") for (f, d) in cur}


@pytest.mark.django_db(transaction=True)
def test_sync_mailmaps_dry_run(postgresql):
    with transaction.atomic():
        populate_mailmap()

    dsn = init_stub_storage_db(postgresql)

    out = call_sync_mailmaps(dsn)

    assert "(dry run)" in out
    assert "Synced 1 mailmaps to swh.storage database" in out

    assert get_displaynames(postgresql) == {}

    assert (
        UserMailmap.objects.filter(
            from_email_verified=True,
            display_name_activated=True,
            mailmap_last_processing_date__isnull=False,
        ).count()
        == 0
    )


@pytest.mark.django_db(transaction=True)
def test_sync_mailmaps_perform(postgresql):
    with transaction.atomic():
        populate_mailmap()

    dsn = init_stub_storage_db(postgresql)

    out = call_sync_mailmaps("--perform", dsn)

    assert "(dry run)" not in out
    assert "Synced 1 mailmaps to swh.storage database" in out

    expected_displaynames = {
        "Original Name V A <from_emailVA@example.com>": "Display Name V A",
        "Original Name V A 2 <from_emailVA@example.com>": "Display Name V A",
        "Original Name V A 3 <from_emailVA@example.com>": "Display Name V A",
    }

    assert get_displaynames(postgresql) == expected_displaynames

    assert (
        UserMailmap.objects.filter(
            from_email_verified=True,
            display_name_activated=True,
            mailmap_last_processing_date__isnull=False,
        ).count()
        == 1
    )


@pytest.mark.django_db(transaction=True)
def test_sync_mailmaps_with_to_email(postgresql):
    with transaction.atomic():
        populate_mailmap()

    dsn = init_stub_storage_db(postgresql)

    call_sync_mailmaps("--perform", dsn)

    expected_displaynames = {
        "Original Name V A <from_emailVA@example.com>": "Display Name V A",
        "Original Name V A 2 <from_emailVA@example.com>": "Display Name V A",
        "Original Name V A 3 <from_emailVA@example.com>": "Display Name V A",
    }

    assert get_displaynames(postgresql) == expected_displaynames

    # Add a non-valid to_email
    with transaction.atomic():
        for mailmap in UserMailmap.objects.filter(
            from_email_verified=True, display_name_activated=True
        ):
            mailmap.to_email = "to_email@example.com"
            mailmap.save()

    call_sync_mailmaps("--perform", dsn)

    assert get_displaynames(postgresql) == expected_displaynames

    # Verify the relevant to_email
    with transaction.atomic():
        for mailmap in UserMailmap.objects.filter(
            from_email_verified=True, display_name_activated=True
        ):
            mailmap.to_email_verified = True
            mailmap.save()

    call_sync_mailmaps("--perform", dsn)

    new_displayname = "Display Name V A <to_email@example.com>"
    expected_displaynames = {
        "Original Name V A <from_emailVA@example.com>": new_displayname,
        "Original Name V A 2 <from_emailVA@example.com>": new_displayname,
        "Original Name V A 3 <from_emailVA@example.com>": new_displayname,
    }

    assert get_displaynames(postgresql) == expected_displaynames


@pytest.mark.django_db(transaction=True)
def test_sync_mailmaps_disable(postgresql):
    """Check that disabling a mailmap only happens once"""
    with transaction.atomic():
        populate_mailmap()

    dsn = init_stub_storage_db(postgresql)

    # Do the initial mailmap sync
    call_sync_mailmaps("--perform", dsn)

    assert len(get_displaynames(postgresql)) == 3

    updated = 0
    # Disable a display name
    with transaction.atomic():
        # Cannot use update() because `last_update_date` would not be updated
        for mailmap in UserMailmap.objects.filter(
            from_email_verified=True, display_name_activated=True
        ):
            mailmap.display_name_activated = False

            mailmap.save()
            updated += 1

    assert updated == 1

    # Sync mailmaps again
    out = call_sync_mailmaps("--perform", dsn)
    assert "1 mailmaps to disable" in out

    assert get_displaynames(postgresql) == {}

    # Update a displayname by hand
    with postgresql.cursor() as cur:
        cur.execute(
            "UPDATE person SET displayname='Manual Display Name' "
            "WHERE fullname='Original Name V A <from_emailVA@example.com>'"
        )

    expected_displaynames = {
        "Original Name V A <from_emailVA@example.com>": "Manual Display Name"
    }

    assert get_displaynames(postgresql) == expected_displaynames

    # Sync mailmaps one last time. No mailmaps should be disabled
    out = call_sync_mailmaps("--perform", dsn)
    assert "0 mailmaps to disable" in out

    assert get_displaynames(postgresql) == expected_displaynames
